import http.client
import urllib.parse
import re
import time

from bloonix.validator import validate
from bloonix.logger import logger
from bloonix.json import json

class REST:
    def __init__(self, **kwargs):
        self.config = self.validate(kwargs)
        self.log = logger("bloonix")
        self.conn = None

        if self.config["proto"] == "https":
            self.httpclient = http.client.HTTPSConnection
            if self.config["port"] == 0:
                self.config["port"] = 443
        else:
            self.httpclient = http.client.HTTPConnection
            if self.config["port"] == 0:
                self.config["port"] = 80

    def get(self, path, **kwargs):
        return self.request("GET", path, **kwargs)

    def put(self, path, **kwargs):
        return self.request("PUT", path, **kwargs)

    def post(self, path, **kwargs):
        """
            self.post("/foo", body='{"foo":"bar"}')
        Or
            self.post("/foo", data={"foo":"bar"})
        """
        return self.request("POST", path, **kwargs)

    def delete(self, path, **kwargs):
        return self.request("DELETE", path, **kwargs)

    def request(self, method, path, status_ok=None, **kwargs):
        data = None
        if "data" in kwargs:
            data = kwargs["data"]
            del kwargs["data"]

        headers = {"Content-Type": "application/json"}
        headers.update(self.config["headers"])

        if "headers" in kwargs:
            headers.update(kwargs["headers"])

        kwargs["headers"] = headers

        if "query_string" in kwargs:
            params = urllib.parse.urlencode(kwargs["query_string"])
            path = "%s?%s" % (path, params)
            del kwargs["query_string"]

        if data != None:
            kwargs["body"] = data
            if isinstance(kwargs["body"], (dict,list)):
                kwargs["body"] = json.dumps(kwargs["body"])
            kwargs["body"] = bytes(kwargs["body"], "utf-8")

        if self.config["autodie"] is True:
            return self._request(method, path, status_ok, kwargs)

        return self._eval_request(method, path, status_ok, kwargs)

    def close(self):
        self.conn.close()
        self.conn = None

    def _eval_request(self, method, path, status_ok, kwargs):
        try:
            return self._request(method, path, status_ok, kwargs)
        except Exception as e:
            return self._errstr(e)

    def _request(self, method, path, status_ok, kwargs):
        if self.conn is None or self.config["autoclose"] is True:
            self.conn = self.httpclient(
                self.config["host"],
                self.config["port"],
                timeout=self.config["timeout"]
            )

        self.conn.request(method, path, **kwargs)
        res = self.conn.getresponse()

        if status_ok is not None:
            if isinstance(status_ok, int):
                status_ok = [status_ok]
            if isinstance(status_ok, str) and re.search(r"^\d\d\d\Z", status_ok):
                status_ok = [int(status_ok)]
            if not isinstance(status_ok, list):
                raise Exception("invalid value for status_ok")
        else:
            status_ok = self.config["status_ok"]

        if res.status not in status_ok:
            raise Exception(
                "request failed '%s:%s%s': [%s %s]" % (
                    self.config["host"], self.config["port"],
                    path, res.status, res.reason
                )
            )

        data = res.read()
        data = data.decode("utf-8")
        data = json.loads(data)

        if self.config["autoclose"] is True:
            self.close()

        return data

    def _errstr(self, msg):
        self.errstr = msg
        return None

    def set_header(self, key, value):
        self.config["headers"][key] = value

    def validate(self, kwargs):
        return validate(kwargs, {
            "proto": {
                "type": str,
                "regex": r"^(http|https)\Z",
                "default": "http"
            },
            "host": {
                "type": str,
                "default": "127.0.0.1"
            },
            "port": {
                "type": int,
                "default": 0
            },
            "timeout": {
                "type": int,
                "default": 60
            },
            "status_ok": {
                "type": list,
                "default": [200]
            },
            "mode": {
                "type": str,
                "regex": r"^(balanced|failover)\Z",
                "default": "balanced"
            },
            "headers": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            "autodie": {
                "type": bool,
                "default": False
            },
            "autoclose": {
                "type": bool,
                "default": True
            }
        })

#rest = REST(proto="http", host="127.0.0.1", timeout="5")
#data = rest.get("/")
#
#if data is None:
#    print(rest.errstr)
#
#print(data)
